/*
 * absolute.c - Main absolute driver definitions
 * Copyright (C) 2005 Ian Gulliver
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <sys/types.h>
#include <sys/kd.h>
#include "xf86Version.h"
#include "misc.h"
#include "xf86.h"
#include "xf86_ansic.h"
#include "xf86_OSproc.h"
#include "xf86Xinput.h"
#include "exevents.h"
#include "mipointer.h"

#include "absolute.h"
#include "convert.h"
#include "linux_usb_hid.h"
#include "linux_event.h"
#include "linux_event_elo.h"
#include "magictouch_serial.h"

#ifdef XFree86LOADER
#include "xf86Module.h"
#endif

static void beep(struct devrec *priv) {
	if (priv->tty_fd == -1)
		return;
	if (ioctl(priv->tty_fd,KIOCSOUND,(char *)5965) != 0)
		return;
	usleep(10000);
	ioctl(priv->tty_fd,KIOCSOUND,(char *)0);
}

static int reload_conffile(struct devrec *priv) {
	if (priv->points != NULL)
		free_points(priv->points);
	if (priv->mappings != NULL)
		free_mappings(priv->mappings);

	priv->points = load_points(priv->calibration_file);
	priv->mappings = build_mappings(priv->points);
	if (priv->mappings == NULL) {
		xf86Msg(X_ERROR, "absolute: unable to build calibration mappings (not enough points?)\n");
		return 1;
	}

	return 0;
}

static int process_button(struct devrec *priv, int *button) {
	int i, c;

	if (*button < 0 || *button > 7)
		return 1;
	if (priv->ignorebutton & (0x01 << *button))
		return 1;
	if (priv->discardbutton & (0x01 << *button))
		return 1;

	for (i = 0, c = 0; i < *button; i++) {
		if (priv->discardbutton & (0x01 << i))
			c++;
	}

	*button -= c;

	for (i = 0; i <= *button; i++) {
		if (priv->fakebutton & (0x01 << i))
			(*button)++;
	}

	return 0;
}

void absolute_button_down(struct devrec *priv, int button) {
	if (process_button(priv,&button) != 0)
		return; /* ignore event */
	priv->button |= (0x01 << button);
}

void absolute_button_up(struct devrec *priv, int button) {
	if (process_button(priv,&button) != 0)
		return; /* ignore event */
	priv->button &= ~(0x01 << button);
}

void absolute_move_x(struct devrec *priv, int_least32_t x) {
	if (priv->switchxy)
		priv->y = x;
	else
		priv->x = x;
}

void absolute_move_y(struct devrec *priv, int_least32_t y) {
	if (priv->switchxy)
		priv->x = y;
	else
		priv->y = y;
}

void absolute_move_z(struct devrec *priv, int_least32_t z) {
	if (z < priv->minZ)
		z = priv->minZ;
	if (z > priv->maxX)
		z = priv->maxZ;
	priv->z = z;
}

static void absolute_readinput(LocalDevicePtr local) {
	struct devrec *priv = local->private;
	int_least32_t oldX = priv->x, oldY = priv->y, oldZ = priv->z;
	uint_least8_t button = priv->button;
	int i;

	priv->parked = 0;

	priv->read_function(local->fd,priv);

	if (priv->allowdrag) {
		if (oldX != priv->x || oldY != priv->y || oldZ != priv->z)
			xf86PostMotionEvent(local->dev, 1, 0, 3, priv->x, priv->y, priv->z);

		if ((button != priv->button) ||
			(oldZ < priv->minButtonZ && priv->z >= priv->minButtonZ) ||
			(oldZ >= priv->minButtonZ && priv->z < priv->minButtonZ)) {
			if (oldZ < priv->minButtonZ && priv->z >= priv->minButtonZ) 
				xf86PostButtonEvent(local->dev, 1, priv->buttonZ, 1, 0, 3, priv->x, priv->y, priv->z);
			if (oldZ >= priv->minButtonZ && priv->z < priv->minButtonZ)
				xf86PostButtonEvent(local->dev, 1, priv->buttonZ, 0, 0, 3, priv->x, priv->y, priv->z);
			for (i = 0; i < 8; i++) {
				uint_least8_t mask = (0x01 << i);
				if ((button & mask) != (priv->button & mask)) {
					/* change in this button */
					xf86PostButtonEvent(local->dev, 1, i+1, priv->button & mask, 0, 3, priv->x, priv->y, priv->z);
				}
			}

			if (priv->button == 0 && priv->z < priv->minButtonZ)
				beep(priv);
			if (priv->park && (priv->button == 0 && priv->z < priv->minButtonZ)) {
				priv->parked = 1;
				xf86PostMotionEvent(local->dev, 1, 0, 3, priv->x, priv->y, priv->z);
			}
		}
	} else if ((button != priv->button) ||
			(oldZ >= priv->minButtonZ && priv->z < priv->minButtonZ)) {
		int didsomething = 0;
		if (oldZ >= priv->minButtonZ && priv->z < priv->minButtonZ) {
			didsomething = 1;
			xf86PostMotionEvent(local->dev, 1, 0, 3, priv->x, priv->y, priv->z);
			xf86PostButtonEvent(local->dev, 1, priv->buttonZ, 1, 0, 3, priv->x, priv->y, priv->z);
			xf86PostButtonEvent(local->dev, 1, priv->buttonZ, 0, 0, 3, priv->x, priv->y, priv->z);
		}
		for (i = 0; i < 8; i++) {
			uint_least8_t mask = (0x01 << i);
			if ((button & mask) && !(priv->button & mask)) {
				didsomething = 1;
				xf86PostMotionEvent(local->dev, 1, 0, 3, priv->x, priv->y, priv->z);
				xf86PostButtonEvent(local->dev, 1, i+1, 1, 0, 3, priv->x, priv->y, priv->z);
				xf86PostButtonEvent(local->dev, 1, i+1, 0, 0, 3, priv->x, priv->y, priv->z);
			}
		}
		if (didsomething) {
			beep(priv);
			if (priv->park) {
				priv->parked = 1;
				xf86PostMotionEvent(local->dev, 1, 0, 3, priv->x, priv->y, priv->z);
			}
		}
	}
}

static int absolute_connect(DeviceIntPtr devptr) {
	LocalDevicePtr local = (LocalDevicePtr) devptr->public.devicePrivate;
	struct devrec *priv = local->private;
	char *split = strchr(priv->device,':');

	if (split == NULL) {
		local->fd = xf86OpenSerial(local->options);
	} else {
		/* multiple devices */
		char *localdev = xf86FindOptionValue(local->options, "Device");
		char *mydev = strdup(priv->device);
		char *tok = mydev;
		char *curdev;

		while ((curdev = strtok(tok,":")) != NULL) {
			tok = NULL;
			strcpy(localdev,curdev);
			local->fd = xf86OpenSerial(local->options);
			if (local->fd != -1)
				break;
		}

		free(mydev);
	}
	InitValuatorAxisStruct(devptr, 0, priv->minX, priv->maxX, priv->maxX / 7.5, 0, priv->maxX / 7.5);
	InitValuatorAxisStruct(devptr, 1, priv->minY, priv->maxY, priv->maxY / 5.5, 0, priv->maxY / 5.5);
	InitValuatorAxisStruct(devptr, 2, priv->minZ, priv->maxZ, 128, 0, 128);
	if (local->fd != -1 && priv->init_function != NULL)
		priv->init_function(local->fd,priv);
	return (local->fd > 0);
}

static Bool absolute_convert(LocalDevicePtr local, int first, int num, int v0, int v1, int v2, int v3, int v4, int v5, int *x, int *y) {
	struct devrec *priv = local->private;
	struct point point;

	if (miPointerCurrentScreen() == NULL)
		return FALSE;

	if (priv->maxX == -1)
		priv->maxX = miPointerCurrentScreen()->width;
	if (priv->maxY == -1)
		priv->maxY = miPointerCurrentScreen()->width;

	if (priv->parked == 1) {
		*x = priv->parkX;
		*y = priv->parkY;
	} else {
		point.Ax = v0;
		point.Ay = v1;

		convert(priv->mappings,&point);

		if (point.Bx < priv->minX)
			point.Bx = priv->minX;
		if (point.Bx > priv->maxX)
			point.Bx = priv->maxX;
		if (point.By < priv->minY)
			point.By = priv->minY;
		if (point.By > priv->maxY)
			point.By = priv->maxY;

		*x = point.Bx;
		*y = point.By;
	}

	xf86XInputSetScreen(local, 0, *x, *y);
	return TRUE;
}

static void absolute_controlproc(DeviceIntPtr devptr, PtrCtrl * ctrl) {
	return;
}

static int absolute_proc(DeviceIntPtr devptr, int operation) {
	LocalDevicePtr local = (LocalDevicePtr) devptr->public.devicePrivate;
	int nbaxes = 3;			/* X Y Pressure */
	int nbuttons = 8;			/* This this is necessary for most apps to work. */
	CARD8 map[8] = { 0, 1, 2, 3, 4, 5, 6, 7 };

	switch (operation) {
		case DEVICE_INIT:
			if (InitButtonClassDeviceStruct(devptr, nbuttons, map) == FALSE)
				return !Success;
			if (InitFocusClassDeviceStruct(devptr) == FALSE)
				return !Success;
			if (InitPtrFeedbackClassDeviceStruct(devptr, absolute_controlproc) == FALSE)
				return !Success;
			if (InitProximityClassDeviceStruct(devptr) == FALSE)
				return !Success;
			if (InitValuatorClassDeviceStruct(devptr, nbaxes, xf86GetMotionEvents, local->history_size, Absolute | OutOfProximity) == FALSE)
				return !Success;
			xf86MotionHistoryAllocate(local);
			absolute_connect(devptr);
			break;
		case DEVICE_ON:
			if (local->fd == -1)
				absolute_connect(devptr);
			if (local->fd == -1)
				return !Success;
			xf86AddEnabledDevice(local);
			devptr->public.on = TRUE;
			break;
		case DEVICE_OFF:
			if (local->fd != -1)
	 			xf86RemoveEnabledDevice(local);
			/* fallthrough */
		case DEVICE_CLOSE:
			if (local->fd != -1) {
				close(local->fd);
				local->fd = -1;
			}
			break;
	}
	return Success;
}

static int absolute_changecontrol(LocalDevicePtr local, xDeviceCtl * control) {
	return Success;
}

static int absolute_switchmode(ClientPtr client, DeviceIntPtr devptr, int mode) {
	LocalDevicePtr local = (LocalDevicePtr) devptr->public.devicePrivate;
	struct devrec *priv = local->private;

	switch (mode) {
		case 3:
			reload_conffile(priv);
			break;
	}

	return Success;
}

static LocalDevicePtr absolute_allocate(InputDriverPtr drv) {
	LocalDevicePtr local = xf86AllocateInput(drv, 0);
	struct devrec *priv = xalloc(sizeof(*priv));

	local->name = "absolute";
	local->flags = 0;
	local->device_control = absolute_proc;
	local->read_input = absolute_readinput;
	local->close_proc = NULL;
	local->control_proc = absolute_changecontrol;
	local->switch_mode = absolute_switchmode;
	local->conversion_proc = absolute_convert;
	local->fd = -1;
	local->atom = 0;
	local->dev = NULL;
	local->private = priv;
	local->type_name = XI_TOUCHSCREEN;
	local->history_size = 0;
	local->old_x = local->old_y = -1;

	priv->device = NULL;
	priv->calibration_file = NULL;
	priv->subdriver = UNKNOWN;
	priv->points = NULL;
	priv->mappings = NULL;
	priv->read_function = NULL;
	priv->init_function = NULL;
	priv->minX = priv->maxX = priv->minY = priv->maxY = priv->minZ = priv->maxZ = 0;
	priv->minButtonZ = 1;
	priv->park = 0;
	priv->parked = 0;
	priv->parkX = priv->parkY = 0;
	priv->ignorebutton = 0;
	priv->discardbutton = 0;
	priv->fakebutton = 0;
	priv->buttonZ = 1;
	priv->switchxy = 0;
	priv->allowdrag = 0;
	priv->x = priv->y = priv->z = 0;
	priv->button = 0;
	priv->tty_fd = -1;

	return local;
}

static void absolute_uninit(InputDriverPtr drv, LocalDevicePtr local, int flags) {
	struct devrec *priv = local->private;

	absolute_proc(local->dev, DEVICE_CLOSE);
	xfree(priv);
	xf86DeleteInput(local, 0);
}

static InputInfoPtr absolute_init(InputDriverPtr drv, IDevPtr dev, int flags) {
	LocalDevicePtr local = NULL;
	struct devrec *priv = NULL;
	pointer options;
	int_least32_t parkX;
	char *subdriver;
	char *devicetype;

	if ((local = absolute_allocate(drv)) == NULL)
		return NULL;
	if (local->private == NULL) {
		xfree(local);
		return NULL;
	}
	local->conf_idev = dev;
	xf86CollectInputOptions(local, NULL, NULL);
	options = local->options;
	local->name = dev->identifier;
	priv = (struct devrec *) local->private;
	priv->device = xf86FindOptionValue(options, "Device");
	if (priv->device == NULL) {
      		xf86Msg(X_ERROR, "absolute: no Device specified\n");
		return NULL;
	}
	priv->device = strdup(priv->device);
	xf86ProcessCommonOptions(local, local->options);

	subdriver = xf86FindOptionValue(options, "SubDriver");
	if (subdriver == NULL) {
      		xf86Msg(X_ERROR, "absolute: no SubDriver specified\n");
		return NULL;
	}
	if (strcasecmp(subdriver,"Linux USB HID") == 0) {
		priv->read_function = linux_usb_hid_read;
	} else if (strcasecmp(subdriver,"Linux Input Event") == 0) {
		priv->read_function = linux_event_read;
	} else if (strcasecmp(subdriver,"Linux Input Event (ELO)") == 0) {
		priv->read_function = linux_event_elo_read;
	} else if (strcasecmp(subdriver,"MagicTouch Serial") == 0) {
		priv->init_function = magictouch_serial_init;
		priv->read_function = magictouch_serial_read;
	} else {
      		xf86Msg(X_ERROR, "absolute: unknown SubDriver specified\n");
		return NULL;
	}


	devicetype = xf86FindOptionValue(options, "DeviceType");
	if (devicetype == NULL || strcasecmp(devicetype,"Touchscreen") == 0)
		local->type_name = XI_TOUCHSCREEN;
	else if (strcasecmp(devicetype,"Tablet") == 0)
		local->type_name = XI_TABLET;
	else {
      		xf86Msg(X_ERROR, "absolute: invalid DeviceType pecified\n");
		return NULL;
	}

	priv->calibration_file = xf86FindOptionValue(options, "CalibrationFile");
	if (priv->calibration_file == NULL) {
      		xf86Msg(X_ERROR, "absolute: no CalibrationFile specified\n");
		return NULL;
	}
	local->name = malloc(strlen(local->type_name)+1+strlen(priv->calibration_file));
	sprintf(local->name,"%s/%s",local->type_name,priv->calibration_file);

	if (reload_conffile(priv) != 0)
		return NULL;

	priv->minX = xf86SetIntOption(options, "MinX", 0);
	priv->minY = xf86SetIntOption(options, "MinY", 0);
	priv->maxX = xf86SetIntOption(options, "MaxX", -1);
	priv->maxY = xf86SetIntOption(options, "MaxY", -1);
	priv->minZ = xf86SetIntOption(options, "MinZ", 0);
	priv->maxZ = xf86SetIntOption(options, "MaxZ", 128); /* this seems to be the standard range for pressure */
	priv->minButtonZ = xf86SetIntOption(options, "MinButtonZ", 1);

	priv->ignorebutton = xf86SetIntOption(options, "IgnoreButton", 0);
	priv->discardbutton = xf86SetIntOption(options, "DiscardButton", 0);
	priv->fakebutton = xf86SetIntOption(options, "FakeButton", 0);
	priv->buttonZ = xf86SetIntOption(options, "ButtonZ", 1);

	priv->switchxy = xf86SetIntOption(options, "SwitchXY", 0);

	priv->allowdrag = xf86SetIntOption(options, "AllowDrag", 0);

	parkX = xf86SetIntOption(options, "ParkX", -1);
	if (parkX == -1)
		priv->park = 0;
	else {
		priv->park = 1;
		priv->parkX = parkX;
		priv->parkY = xf86SetIntOption(options, "ParkY", -1);
		if (priv->parkY == -1) {
      			xf86Msg(X_ERROR, "absolute: ParkX without ParkY\n");
			return NULL;
		}
	}

	if (xf86SetIntOption(options, "Beep", 0) == 1) {
		priv->tty_fd = open("/dev/tty0",O_WRONLY);
	} else
		priv->tty_fd = -1;

	local->flags |= XI86_POINTER_CAPABLE | XI86_CONFIGURED;

	return (local);
}

InputDriverRec absolute = { 1, "absolute", NULL, absolute_init, absolute_uninit, NULL, 0 };

#ifdef XFree86LOADER

static void absolute_unplug(pointer p) {
	return;
}

static pointer absolute_plug(pointer module, pointer options, int *errmaj, int *errmin) {
	xf86AddInputDriver(&absolute, module, 0);
	return module;
}

static XF86ModuleVersionInfo absolute_versionrec = {
	"absolute",
	MODULEVENDORSTRING,
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	1, 0, 0,
	ABI_CLASS_XINPUT,
	ABI_XINPUT_VERSION,
	MOD_CLASS_XINPUT,
	{0, 0, 0, 0}		/* signature, to be patched into the file by a tool */
};

XF86ModuleData absoluteModuleData = {
	&absolute_versionrec,
	absolute_plug,
	absolute_unplug
};
#endif
