/*
 * convert.c - Coordinate system conversion definitions
 * Copyright (C) 2005 Ian Gulliver
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "xf86.h"
#include "convert.h"

struct mapping_result {
	int_least64_t weakness;
	int_least32_t Bx;
	int_least32_t By;
	double value;
	struct strength *next;
};

struct point *add_point(struct point *old, int_least32_t Ax, int_least32_t Ay, int_least32_t Bx, int_least32_t By) {
	struct point *new = malloc(sizeof(*new));
	new->next = old;

	new->Ax = Ax;
	new->Ay = Ay;
	new->Bx = Bx;
	new->By = By;

	return new;
}

struct mapping *build_mappings(struct point *head) {
	struct mapping *old = NULL;
	struct point *iter1;
	uint_least16_t id = 0;

	/* these are complete BS, but they keep us from failing to start and preventing calibration */
	if (head == NULL)
		head = add_point(head,0,0,0,0);
	if (head->next == NULL)
		head = add_point(head,1,1,1,1);

	iter1 = head;

	while (iter1 != NULL) {
		struct point *iter2 = iter1->next;
		while (iter2 != NULL) {
			/* every combination, ignoring order */
			struct mapping *new;
	
			if ((iter1->Bx == iter2->Bx) ||
				(iter1->By == iter2->By) ||
				(iter1->Ax == iter2->Ax) ||
				(iter1->Ay == iter2->Ay)) {
				/* points on the same line */
				iter2 = iter2->next;
				continue;
			}
	
			new = malloc(sizeof(*new));
			new->id = id++;
			new->next = old;
			new->point1 = iter1;
			new->point2 = iter2;

			new->X_scale = (double) (iter2->Bx - iter1->Bx) / (iter2->Ax - iter1->Ax);
			new->X_offset = iter1->Bx - (iter1->Ax * new->X_scale);
			new->Y_scale = (double) (iter2->By - iter1->By) / (iter2->Ay - iter1->Ay);
			new->Y_offset = iter1->By - (iter1->Ay * new->Y_scale);

			old = new;
			iter2 = iter2->next;
		}
		iter1 = iter1->next;
	}

	return old;
}

static int_least64_t distance(int_least32_t Ax, int_least32_t Ay, int_least32_t Bx, int_least32_t By) {
	int_least32_t Xdiff, Ydiff;

	Xdiff = Bx - Ax;
	Ydiff = By - Ay;
	return sqrt((Xdiff * Xdiff) + (Ydiff * Ydiff));
}

void convert(struct mapping *mappings, struct point *point) {
	struct mapping_result *results = malloc(sizeof(*results) * (mappings->id + 1));
	struct mapping *iter = mappings;
	uint_least64_t max_weakness = 0;
	uint_least64_t min_weakness = UINT64_C(1) << 63;
	int_least64_t Bx = 0, By = 0;
	double total_value = 0;
	int i = 0;

	while (iter != NULL) {
		results[i].weakness = distance(point->Ax,point->Ay,iter->point1->Ax,iter->point1->Ay),
			 + distance(point->Ax,point->Ay,iter->point2->Ax,iter->point2->Ay);
		results[i].weakness *= results[i].weakness; /* make the falloff exponential */

		if (results[i].weakness > max_weakness)
			max_weakness = results[i].weakness;
		if (results[i].weakness < min_weakness)
			min_weakness = results[i].weakness;

		results[i].Bx = (double) (point->Ax * iter->X_scale) + iter->X_offset;
		results[i].By = (double) (point->Ay * iter->Y_scale) + iter->Y_offset;

		iter = iter->next;
		i++;
	}

	/*
	 * this algorithm essentially disregards the worst case
	 * if everything gets close to worst case, we get huge rounding error because everything sucks
	 * if it is, fake it by making them all at least 50% valuable
	 */

	if (min_weakness * 2 > max_weakness)
		max_weakness = min_weakness * 2;

	/* pass 2 */
	for (i = 0; i <= mappings->id; i++) {
		results[i].value = (double) (max_weakness - results[i].weakness) / max_weakness;

		total_value += results[i].value;

		Bx += (double) results[i].Bx * results[i].value;
		By += (double) results[i].By * results[i].value;
	}

	free(results);

	point->Bx = (double) Bx / total_value;
	point->By = (double) By / total_value;
}

void free_points(struct point *iter) {
	struct point *next;

	while (iter != NULL) {
		next = iter->next;
		free(iter);
		iter = next;
	}
}

void free_mappings(struct mapping *iter) {
	struct mapping *next;

	while (iter != NULL) {
		next = iter->next;
		free(iter);
		iter = next;
	}
}

struct point *load_points(char *filename) {
	struct point *ret = NULL;
	FILE *f;
	char buf[256];

	f = fopen(filename,"r");
	if (f == NULL)
		return NULL;

	while (fgets(buf,256,f) != NULL) {
		char *tbuf = buf;
		char *tok;
		int i = 0;
		int_least32_t coord[4];
		int l = strlen(buf);

		if (l == 0 || buf[0] == '#')
			continue;

		buf[l] = '\0';
		
		while ((tok = strtok(tbuf,",")) != NULL) {
			tbuf = NULL;
			if (i > 3)
				return NULL;
			coord[i++] = atol(tok);
		}

		if (i != 4)
			return NULL;

		ret = add_point(ret,coord[0],coord[1],coord[2],coord[3]);
	}

	return ret;
}
